﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using SimpleJSON;

namespace JustAnotherUser {
    public class ItemDuplicator : MVRScript {
        protected class AtomData {
            private string _id, _type;
            private IDictionary<string, JSONClass> _storables;

            public string id { get { return this._id; } }
            public string type { get { return this._type; } }
            public IDictionary<string, JSONClass> storables { get { return this._storables; } }

            public AtomData(Atom a) {
                this._id = a.uid;
                this._type = a.type;

                this._storables = new Dictionary<string, JSONClass>();
                foreach (JSONClass e in ItemDuplicator.GetStorablesJsonFromSave(this._id)) {
                    if (!e.HasKey("id")) continue;

                    this._storables.Add(e["id"], e);
                }
            }
        }

        private string _duplicateAtomName;
        private AtomData _duplicateAtomData; // TODO save in storable JSON

        // transform difference between the duplicateAtom and the current object
        private Vector3 _spawnPosition;
        private Vector3 _spawnRotation;

        private float _timeSinceLastCheck = 0f;

        private JSONStorableStringChooser _duplicateAtomStorable;
        private JSONStorableFloat _intervalStorable;

        public override void Init() {
            // plugin VaM GUI description
            pluginLabelJSON.val = "ItemDuplicator v1.0";
        }


        // Runs once when plugin loads (after Init)
        protected void Start() {
            this._duplicateAtomStorable = new JSONStorableStringChooser("duplicate", null, "", "Duplicate", (string colliderName) => {
                this._duplicateAtomName = colliderName;
                Atom duplicateAtom = SuperController.singleton.GetAtomByUid(colliderName);
                this._duplicateAtomData = new AtomData(duplicateAtom);

                // if you+difference must result in final; then final-you is difference
                this._spawnPosition = duplicateAtom.mainController.transform.position - containingAtom.mainController.transform.position;
                this._spawnRotation = duplicateAtom.mainController.transform.rotation.eulerAngles - containingAtom.mainController.transform.rotation.eulerAngles; // TODO la rotación afecta también a la posición

                //SuperController.LogMessage("New collider");
            });
            RegisterStringChooser(this._duplicateAtomStorable);
            var linkPopup = CreateFilterablePopup(this._duplicateAtomStorable);
            linkPopup.popupPanelHeight = 600f;
            linkPopup.popup.onOpenPopupHandlers += () => { this._duplicateAtomStorable.choices = SuperController.singleton.GetAtoms().Select(a => a.name).Distinct().ToList(); };

            this._intervalStorable = new JSONStorableFloat("interval", 1.0f, 0.01f, 20.0f);
            RegisterFloat(this._intervalStorable);
            CreateSlider(this._intervalStorable, true); // create on the right

            var spawn = CreateButton("Spawn now");
            spawn.height = 100;
            spawn.button.onClick.AddListener(() => StartCoroutine(SpawnAtom()));

            //SuperController.LogMessage(pluginLabelJSON.val + " Loaded");

            LoadJson(GetPluginJsonFromSave());
        }

        public IEnumerator SpawnAtom() {
            // does it already exists?
            Atom duplicateAtom = SuperController.singleton.GetAtomByUid(this._duplicateAtomName);
            if (duplicateAtom == null) {
                // no atom; generate a new one
                yield return SuperController.singleton.AddAtomByType(this._duplicateAtomData.type, this._duplicateAtomData.id);
                duplicateAtom = SuperController.singleton.GetAtomByUid(this._duplicateAtomName);

                if (this._duplicateAtomData.type == "CustomUnityAsset" && this._duplicateAtomData.storables["asset"] != null) {
                    // load the asset
                    JSONStorable asset = duplicateAtom.GetStorableByID("asset");
                    JSONClass assetJSON = asset.GetJSON();
                    assetJSON.Add("assetName", this._duplicateAtomData.storables["asset"]["assetName"].Value);
                    assetJSON.Add("assetUrl", this._duplicateAtomData.storables["asset"]["assetUrl"].Value);
                    asset.RestoreFromJSON(assetJSON);
                }

                // plugins?
                JSONStorable plugins;
                if ((plugins = duplicateAtom.GetStorableByID("PluginManager")) != null && this._duplicateAtomData.storables["PluginManager"] != null) {
                    JSONClass pluginsJSON = plugins.GetJSON();
                    foreach (string pluginName in this._duplicateAtomData.storables["PluginManager"]["plugins"].AsObject.Keys) {
                        pluginsJSON["plugins"].Add(pluginName, this._duplicateAtomData.storables["PluginManager"]["plugins"][pluginName].Value);
                    }
                    plugins.LateRestoreFromJSON(pluginsJSON);
                }
                // TODO add the rest of the plugins data
            }

            duplicateAtom.mainController.transform.position = containingAtom.mainController.transform.position + this._spawnPosition;
            duplicateAtom.mainController.transform.rotation = Quaternion.Euler(containingAtom.mainController.transform.rotation.eulerAngles + this._spawnRotation);
        }

        public void Update() {
            this._timeSinceLastCheck += Time.deltaTime;
            if (this._timeSinceLastCheck < this._intervalStorable.val) return;

            // maybe the atom is gone?
            if (SuperController.singleton.GetAtomByUid(this._duplicateAtomName) == null) StartCoroutine(SpawnAtom());
            this._timeSinceLastCheck = 0;
        }

        private void LoadJson(JSONNode node) {
            if (node == null) return;

            foreach (string entry in node.AsObject.Keys) {
                switch (entry) {
                    case "duplicate":
                        this._duplicateAtomStorable.val = node[entry].Value;
                        break;

                    case "interval":
                        this._intervalStorable.val = node[entry].AsFloat;
                        break;

                    case "id":
                    case "pluginLabel":
                    default:
                        break; // ignore
                }
            }
        }

        public JSONNode GetPluginJsonFromSave() {
            JSONArray storables = ItemDuplicator.GetStorablesJsonFromSave(containingAtom.name);
            if (storables == null) return null;

            foreach (JSONNode storable in storables) {
                if (storable["id"].Value == this.storeId) {
                    return storable;
                }
            }
            return null;
        }

        // @author https://raw.githubusercontent.com/ChrisTopherTa54321/VamScripts/master/FloatMultiParamRandomizer.cs
        public static JSONArray GetStorablesJsonFromSave(string id) {
            foreach (JSONNode atoms in SuperController.singleton.loadJson["atoms"].AsArray) {
                if (atoms["id"].Value != id) continue;

                return atoms["storables"]?.AsArray;
            }

            return null;
        }
    }
}